/***************************************************************************
 *   Copyright (C) 2006 by Johnny Andersson                                *
 *   andersson.johnny@gmail.com                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qtextcodec.h>
#include <iostream>
#include <qmap.h>
#include "kanjidic.h"
#include "dictmanager.h"
#include "conv.h"

using std::cerr;
using std::cout;

Conv::Conv(QObject *parent, const char *name)
        : QObject(parent, name)
{}


Conv::~Conv()
{}

Conv::ConvertErrors Conv::convertKanjiDicFile(const QString& sourceFileName, const QString& radFileName, const QString& destFileName, const QString& codecName, const QString& radCodecName)
{
    QFile sourceFile;
    QFile destFile;
    QFile metaFile;
    QString str;
    QTextStream ts;
    QDataStream ds;
    QChar cc;
    QTextCodec *codec = 0;
    QValueList<KanjiTemp> entries;
    QValueList<KanjiTemp>::Iterator it;
    KanjiTemp kt;
    KanjiDic kd;
    QIntDict<QString> kanjiToRadicals;
    QIntDict<QString> radicals;

    cc = '#';                   // don't count lines beginning with cc ("comment character")

    bool ok;

    ok = kd.loadRadicals(radFileName, radCodecName);

    if(!ok)
        return NoRadicalFile;

    if(codecName != "utf8")     // utf8 encoding is supported directly by QTextStream
    {
        codec = QTextCodec::codecForName(codecName);

        if(codec == 0)
            return NoCodec;
    }

    sourceFile.setName(sourceFileName);
    destFile.setName(destFileName+".bin");

    if(sourceFile.open(IO_ReadOnly) == false)
        return NoSourceFile;

    if(destFile.open(IO_WriteOnly) == false)
    {
        sourceFile.close();
        return NoDestFile;
    }

    if(codec)
        ts.setCodec(codec);
    else
        ts.setEncoding(QTextStream::UnicodeUTF8);

    ts.setDevice(&sourceFile);
    ds.setDevice(&destFile);

    cout << "Parsing...\n";

    while(!ts.eof())
    {
        str = ts.readLine();

        if(str.left(1) != cc)
        {
            KanjiTemp tmp;
            tmp = parseKanjiDicLine(str);
            tmp.radicals = kd.getRadicals(QChar(tmp.unicode)).utf8();
            entries.append(tmp);
        }
    }

    sourceFile.close();

    // Write number of entries

    uint tmp = entries.count();
    ds.writeRawBytes((const char *)&tmp, sizeof(tmp));

    // Write the offsets to each structure

    for(it = entries.begin(); it != entries.end(); ++it)
    {
        kt = *it;

        tmp = sizeof(DictKanjiStorage) + kt.kunReadings.size() + kt.onReadings.size() + kt.nameReadings.size() + kt.meanings.size() + kt.radicals.size();

        if((tmp % 4) != 0)
        {
            kt.pad = 4 - (tmp % 4);
            tmp += kt.pad;
        }
        else
            kt.pad = 0;

        ds.writeRawBytes((const char *)&tmp, sizeof(tmp));
        (*it).pad = kt.pad;

    }

    for(it = entries.begin(); it != entries.end(); ++it)
    {
        DictKanjiStorage ks;
        kt = *it;

        // Write KanjiStorage header
        // Adding the size of the fields individually is more reliable than using sizeof(struct) because of
        // compiler padding :(.

        ks.strokes = kt.strokes;
        ks.grade = kt.grade;
        ks.freq = kt.freq;
        ks.unicode = kt.unicode;
        ks.jis = kt.jis;
        ks.kunLen = kt.kunReadings.size();
        ks.onLen = kt.onReadings.size();
        ks.nameLen = kt.nameReadings.size();
        ks.meaningsLen = kt.meanings.size();
        ks.radicalsLen = kt.radicals.size();

        ds.writeRawBytes((const char *)&ks, sizeof(ks));

        // Write actual data

        ds.writeRawBytes(kt.kunReadings.data(), kt.kunReadings.size());
        ds.writeRawBytes(kt.onReadings.data(), kt.onReadings.size());
        ds.writeRawBytes(kt.nameReadings.data(), kt.nameReadings.size());
        ds.writeRawBytes(kt.meanings.data(), kt.meanings.size());
        ds.writeRawBytes(kt.radicals.data(), kt.radicals.size());

        // make sure next data structure begins on a even 4-byte boundary
        // as required by the StrongARM CPU

        uchar dat[kt.pad];
        ds.writeRawBytes((const char *)dat, kt.pad);

    }

    destFile.close();
    entries.clear();


    return OK;
}




KanjiTemp Conv::parseKanjiDicLine(const QString& line)
{
    QString tmp;
    uint pos = 1;
    uint oldpos;
    int right;
    bool ok;
    bool strokesFound = false;
    bool nanori = false;
    KanjiTemp ret;
    QString kunReadings;
    QString onReadings;
    QString meanings;
    QString nameReadings;

    // Read the unicode
    ret.unicode = line[0].unicode();

    // Read the JIS code
    pos = line.find(' ', 2);
    tmp = line.mid(2, (pos-2));
    ret.jis = (ushort) tmp.toInt(&ok,16);

    while(pos < line.length())
    {

        oldpos = pos;

        while(line[pos] == ' ')
            pos++;

        // Nanori marker - remaining kana on this line are name readings
        if(line.mid(pos,2) == "T1" || line.mid(pos,2) == "T2") // T2 means readings as radical, don't care for now
        {
            nanori = true;
            pos += 3;
        }

        // Numerical values
        if((line[pos] == 'S' || line[pos] == 'F' || line[pos] == 'G')
                &&
                (line[pos+1].isDigit()))
        {
            char type = line[pos];

            pos++;
            right = line.find(' ',pos);
            tmp = line.mid(pos, (right-pos));
            pos += (right - pos) + 1;

            if(type == 'S')
            {
                if(!strokesFound) // If we've already found stroke count, this is "common miscount"
                    ret.strokes = (uchar) tmp.toInt(&ok,10);
                strokesFound = true;
            }
            if(type == 'F')
                ret.freq = (ushort) tmp.toInt(&ok,10);

            if(type == 'G')
                ret.grade = (uchar)tmp.toInt(&ok,10);
        }

        // English meaning
        if(line[pos] == '{')
        {
            nanori = false;
            pos++;
            right = line.find('}',pos);
            meanings += line.mid(pos, right-pos) + "|";
            pos += right - pos;
        }

        // Kana readings
        // Name

        if(nanori)
        {
            right = line.find(' ',pos);
            nameReadings += line.mid(pos, right-pos).stripWhiteSpace() + "|";
            pos += right - pos;
        }

        // Kun
        else if(KanjiDic::isHiragana(line[pos]))
        {
            if(line[pos-1] == '-')
                pos--;  // if prefix '-' exists, include it

            right = line.find(' ',pos);

            kunReadings += line.mid(pos, right-pos) + "|";
            pos += right - pos;
        }

        // On
        else if(KanjiDic::isKatakana(line[pos]))
        {
            if(line[pos-1] == '-')
                pos--;  // if prefix '-' exists, include it

            right = line.find(' ',pos);
            onReadings += line.mid(pos, right-pos) + "|";
            pos += right - pos;
        }

        if(pos == oldpos)
            pos++;
    }

    if(meanings != "")
        meanings = meanings.left(meanings.length()-1); // remove last '|'

    if(kunReadings != "")
        kunReadings = kunReadings.left(kunReadings.length()-1);

    if(onReadings != "")
        onReadings = onReadings.left(onReadings.length()-1);

    if(nameReadings != "")
        nameReadings = nameReadings.left(nameReadings.length()-1);

    ret.kunReadings = kunReadings.utf8();
    ret.onReadings = onReadings.utf8();
    ret.nameReadings = nameReadings.utf8();
    ret.meanings = meanings.utf8();


    return ret;

}

Conv::ConvertErrors Conv::readWFreqFile(const QString& fileName)
{
    uint x;
    bool ok;
    QString str;
    QStringList pair;
    QFile file(fileName);
    QTextStream ts(&file);

    ts.setCodec(QTextCodec::codecForName("eucJP"));


    if(file.open(IO_ReadOnly) == false)
        return NoWFreqFile;

    while(!ts.eof())
    {
        str = ts.readLine();
        if(str[0] != '#' && str.length()>1)
        {
            pair = pair.split('\t',str);

            x = pair[1].toUInt(&ok);

            if(x > 10000)
                _wfreq.insert(pair[0], "c1");
            else if(x > 5000)
                _wfreq.insert(pair[0], "c2");
            else if(x > 1000)
                _wfreq.insert(pair[0], "c3");
            else if(x > 500)
                _wfreq.insert(pair[0], "c4");
            else if(x > 100)
                _wfreq.insert(pair[0], "c5");
            else if(x > 50)
                _wfreq.insert(pair[0], "c6");
            else if(x > 20)
                _wfreq.insert(pair[0], "c7");

        }
    }

    cout << _wfreq.count() << endl;

    return Conv::OK;
}

Conv::ConvertErrors Conv::convertFile(const QString& sourceFileName, const QString& destFileName, Dictionary::DictFormats dictFormat, const QString& codecName, const QString& markerFileName)
{
    QFile sourceFile;
    QFile destFile;
    QTextStream ts;
    QDataStream ds;
    QCString cstr;
    QString str;
    uint tmp;
    QValueList<DictEntryTemp> entries;
    QValueList<DictEntryTemp>::Iterator it;
    DictEntryTemp det;
    QTextCodec *codec = 0;
    QChar cc;
    DictManager dm;

    if(dm.loadMarkers(markerFileName) == false)
        return NoMarkerFile;

    if(dictFormat == Dictionary::Edict || dictFormat == Dictionary::Enamdict)
        cc = '#';   // ignore lines beginning with cc ("comment character")

    if(codecName != "utf8")     // utf8 encoding is supported directly by QTextStream
    {
        codec = QTextCodec::codecForName(codecName);

        if(codec == 0)
            return NoCodec;
    }

    sourceFile.setName(sourceFileName);
    destFile.setName(destFileName+".bin");

    if(sourceFile.open(IO_ReadOnly) == false)
        return NoSourceFile;

    if(destFile.open(IO_WriteOnly) == false)
    {
        sourceFile.close();
        return NoDestFile;
    }

    if(codec)
        ts.setCodec(codec);
    else
        ts.setEncoding(QTextStream::UnicodeUTF8);

    ts.setDevice(&sourceFile);
    ds.setDevice(&destFile);
	ds.setByteOrder(QDataStream::LittleEndian);

    int count  = 0;

    while(!ts.eof())
    {
        str = ts.readLine();

        if(str.left(1) != cc)
        {
            switch(dictFormat)
            {
            case Dictionary::Edict:
                if(count % 10000 == 0)
                    cout << "Parsing line " << count << "...\n";
                count++;
                entries.append(parseEdictLine(dm.markers(), str));
                break;
            case Dictionary::Enamdict:
                break;
            }
        }
    }

    cout << "Parsed " << count << " lines\n";

    sourceFile.close();

    cout << "Sorting...\n";

	qHeapSort(entries);

    // Write number of entries

    tmp = entries.count();
    ds.writeRawBytes((const char *)&tmp, sizeof(tmp));

	tmp = 4;
    ds.writeRawBytes((const char *)&tmp, sizeof(tmp));  // wordOffs
	tmp = 0;											// dummy value: rewritten later
    ds.writeRawBytes((const char *)&tmp, sizeof(tmp));	// kanaOffs
	tmp = 0;											// dummy value: rewritten later
    ds.writeRawBytes((const char *)&tmp, sizeof(tmp));	// kanjiOffs

    // Write the offsets to each structure

    for(it = entries.begin(); it != entries.end(); ++it)
    {
        det = *it;

        tmp = sizeof(DictWordStorage) + det.markers.size() + det.jpn.size() + det.reading.size() + det.english.size();

        if((tmp % 4) != 0)
        {
            det.pad = 4 - (tmp % 4);
            tmp += det.pad;
        }
        else
            det.pad = 0;

        ds.writeRawBytes((const char *)&tmp, sizeof(tmp));
        (*it).pad = det.pad;
    }

    for(it = entries.begin(); it != entries.end(); ++it)
    {
        class DictWordStorage des;
        det = *it;

        // Write header (the DictEntryStorage structure)

        des.markersLen = det.markers.size();
        des.jpnLen = det.jpn.size();
        des.readingLen = det.reading.size();
        des.englishLen = det.english.size();

        ds.writeRawBytes((const char *)&des, sizeof(des));

        // Write actual data

        ds.writeRawBytes(det.markers, des.markersLen);
        ds.writeRawBytes(det.jpn.data(), des.jpnLen);
        ds.writeRawBytes(det.reading.data(), des.readingLen);
        ds.writeRawBytes(det.english.data(), des.englishLen);

        // make sure next data structure begins on a even 4-byte boundary
        // as required by the StrongARM CPU

        uchar dat[det.pad];
        ds.writeRawBytes((const char *)dat, det.pad);

    }

    cout << "Creating character index...\n";

	QMemArray<int> arr = createCharIndex(entries);

	uint kanaPos = destFile.at();

	for(int i=0; i<arr.count(); i++)
		ds << arr[i];

	cerr << "Creating kanji index...\n";

	arr = createKanjiIndex(entries);

	uint kanjiPos = destFile.at();
	
	for(int i=0; i<arr.count(); i++)
		ds << arr[i];

	destFile.at(2 * sizeof(uint));
    ds.writeRawBytes((const char *)&kanaPos, sizeof(uint));
    ds.writeRawBytes((const char *)&kanjiPos, sizeof(uint));


	destFile.close();

    entries.clear();


    return OK;
}

DictEntryTemp Conv::parseEdictLine(const QValueList<Marker>& markers, const QString& line)
{
    DictEntryTemp ret;
    QString str;
    QString kanji;
    QString kana;
    QString english;
    QValueList<uchar> wordMarkers;

    int p;
    int p1;
    int p2;

    p = line.find(' ');

    if(line[p+1] == '[')
    {
        kanji = line.left(p);
        p1 = p + 2;                     // index of first kana char
        p2 = line.find(']',p+2);        // index of last kana char
        kana = line.mid(p1, p2-p1);
        english = line.mid(p2+3, line.length() - (p2+3));
    }
    else
    {
        kana = line.left(p);
        english = line.mid(p+2, line.length() - (p+2));
    }

    p1 = 0;


    // Look for markers in english text
    while((p1 = english.find('(', p1)) != -1)
    {
        bool erase = false;
        QString part;
        QStringList strlist;

        p2 = english.find(')',p1+1);
        part = english.mid(p1+1,p2-p1-1);

        if(part.find(',') != -1)
            // the marker may be of type "adj-na,adv,exp,n" - let's split it
            strlist = strlist.split(',',part);
        else
            strlist.append(part);

        // Now let's see if they are markers or just text inside parenthesis
        for(QStringList::Iterator it = strlist.begin(); it != strlist.end(); ++it)
        {
            for(int i = 0; i<markers.count(); i++)
            {
                if(markers[i].marker == (*it))
                {
                    wordMarkers.append((uchar)i);
                    erase = true;
                }
            }
        }

        strlist.clear();

        if(erase)
        {
            int rp1 = p1;
            int rp2 = p2;

            if(rp2 < english.length()-1)
            {
                if(english[rp2+1] == ' ')
                    rp2++;
            }
            english.remove(rp1,rp2-rp1+1);
            erase = false;
        }
        else
            p1 = p2;
    }

    // Add our own markers (for common words)

    QString comm = _wfreq[kanji];
    if(comm.length() == 0)
        comm = _wfreq[kana];

    if(comm.length() > 0)
    {
        for(int i = 0; i<markers.count(); i++)
        {
            if(markers[i].marker == comm)
                wordMarkers.append((uchar)i);
        }
    }
    
    QByteArray wmarkers(wordMarkers.count());
    for(int i=0; i<wordMarkers.count(); i++)
        wmarkers[i] = wordMarkers[i];

    while(english.right(1) == "/" || english.right(1) == " ")
        english.truncate(english.length()-1);

    ret.markers = wmarkers;
    ret.jpn = kanji.utf8();
    ret.reading = kana.utf8();
    ret.english = english.utf8();
    return ret;
}

QMemArray<int> Conv::createCharIndex(QValueList<DictEntryTemp>& entries)
{
    QChar ch;
    int index = 0;
    const DictEntryTemp *de;
    QString reading;
    QString utfchar;
    QMap<QChar, int> firstCharIndex;
    QMap<QChar, int> lastCharIndex;

    while(index < entries.count())
    {

        cerr << int(100 * float(index) / float(entries.count())) << "%\n";

        de = &entries[index];

        utfchar = QString(reading[0]).local8Bit();

        reading = QString::fromUtf8(de->reading);
        ch = reading[0];
        firstCharIndex.insert(ch, index);

        while(reading[0] == ch && index < entries.count())
        {
            de = &entries[index];
            reading = QString::fromUtf8(de->reading);
            index++;
        }

        if(index >= entries.count())
            break;

        lastCharIndex.insert(ch, index - 1);
    }

//    QString ret ="@firstchar_index\n";

	QMemArray<int> ret(firstCharIndex.count() * 3);
	QMap<QChar, int>::Iterator it;
	int pos = 0;
    for( it = firstCharIndex.begin(); it != firstCharIndex.end(); ++it )
	{
		ret[pos++] = it.key().unicode();
		ret[pos++] = firstCharIndex[it.key()];
		ret[pos++] = lastCharIndex[it.key()];
	}

//        ret += QString(it.key()) + "\n" + QString().setNum(firstCharIndex[it.key()]) + "\n" + QString().setNum(lastCharIndex[it.key()]) + "\n";

//    ret += "\n@\n";

    return ret;

}

QMemArray<int> Conv::createKanjiIndex(QValueList<DictEntryTemp>& entries)
{
	QMap< QChar, QValueList<int> > map;
	QMemArray<int> ret;
	QString kanji;
	DictEntryTemp *de;
	int index = 0;
	int count = 0;

    while(index < entries.count())
    {
		count++;
		if((count % 1000) == 0)
		{
			cerr << count << "/" << entries.count() << "\n";
		}

		de = &entries[index];

        kanji = QString::fromUtf8(de->jpn);

		for(int i=0; i<kanji.length(); i++)
		{
			QChar k = kanji[i];
			if(KanjiDic::isKanji(k))
			{
				if(map.find(k) == map.end())
				{
					QValueList<int> vlist;
					vlist.append(index);
					map.insert(k, vlist);
				}
				else
				{
					map[k].append(index);
				}
			}
		}
		index++;
	}

	QMap<QChar, QValueList<int> >::Iterator it;

	cerr << "Number of kanji: " << map.count() << "\n";

	count = 0;

	for(it = map.begin(); it != map.end();  ++it)
	{
		count += it.data().count();
//		cerr << "\n" << QString(it.key()).utf8() << ": ";
//		for(int x=0; x < it.data().count(); x++)
//			 cerr << it.data()[x] << ",";
	}

	cerr << "Number of kanji pointers: " << count << "\n";

	ret.resize(map.count()*2 + count);

	int offset = map.count()*2;
	int kanjipos = 0;

	for(it = map.begin(); it != map.end();  ++it)
	{
		ret[kanjipos] = it.key().unicode();
		kanjipos++;
		ret[kanjipos] = offset;
		kanjipos++;

		for(int i=0; i<it.data().count(); i++)
			ret[offset++] = it.data()[i];
	}



	return ret;

}


