/*
    JARDS, Japanese Reference and Drilling System.
    Copyright (C) 2006  Johnny Andersson (e-mail: andersson.johnny@gmail.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/mman.h>
#include <iostream>
#include <qtextcodec.h>
#include <qtextstream.h>
#include <qdatastream.h>
#include <qstringlist.h>
#include <qtextbrowser.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qregexp.h>
#include "dictionary.h"
#include "cstring.h"
#include "kanjidic.h"

using namespace std;


DictWord::DictWord()
{}

QByteArray DictWord::markers() const
{
    return QByteArray().duplicate((const char *) uint(_ptr) + sizeof(DictWordStorage), _ptr->markersLen);
}

const CString DictWord::jpn() const
{
    return CString((const char *) (uint(_ptr) + sizeof(DictWordStorage) + _ptr->markersLen), _ptr->jpnLen);
}

const CString DictWord::reading() const
{
    return CString((const char *) (uint(_ptr) + sizeof(DictWordStorage) + _ptr->markersLen + _ptr->jpnLen), _ptr->readingLen);
}

const CString DictWord::english() const
{
    return CString((const char *) (uint(_ptr) + sizeof(DictWordStorage) + _ptr->markersLen + _ptr->jpnLen + _ptr->readingLen), _ptr->englishLen);
}

QString DictWord::infoString() const
{
    QString ret;

    ret = QString::fromUtf8(jpn()) + " [" + QString::fromUtf8(reading()) + "] " + QString::fromUtf8(english());

    return ret;

}

Dictionary::Dictionary(const QValueList<Marker>& markers)
{
    _entries = 0;
    _file._mmapBase = 0;
    _markers = markers;
    _cancelled=false;
    _searchMode = Exact;

}

Dictionary::Dictionary(const QValueList<Marker>& markers, const QString& fileName)
{
    _entries = 0;
    _file._mmapBase = 0;
    _markers = markers;
    _cancelled=false;
    _searchMode = Exact;
    setFile(fileName);
}

Dictionary::~Dictionary()
{
    if(_entries != 0)
        delete[] _entries;

    if(_file._qfile.isOpen())
    {
        munmap(_file._mmapBase, _file._qfile.size());
        _file._qfile.close();
    }

}

void Dictionary::getKanaRange(QChar ch, int *first, int *last) const
{
	uint *kanaIndexBase = (uint *)_file._mmapBase;
	kanaIndexBase += _file._kanaOffs / sizeof(uint);
	int kanaOffs = 0;

	*first = -1;
	*last = -1;

	while(1)
	{
		if(kanaIndexBase[kanaOffs] == ch.unicode())
		{
			*first = kanaIndexBase[kanaOffs+1];
			*last = kanaIndexBase[kanaOffs+2];
			return;
		}
		kanaOffs += 3;
		if(kanaOffs >= _file._kanaIndexSize)
			return;
	}

}

bool Dictionary::setFile(const QString& fileName)
{
    uint offset;

    _file._qfile.setName(fileName);

    if(!_file._qfile.open(IO_ReadOnly))
    {
        cerr << "failed to open " << fileName << endl;
        return false;
    }

    _file._mmapBase = mmap(0, _file._qfile.size(), PROT_READ, MAP_SHARED, _file._qfile.handle(), 0);

    if(_file._mmapBase == MAP_FAILED)
    {
        _file._mmapBase = 0;
        _file._qfile.close();
        cerr << "mmap failed in dict::setFile" << endl;
        return false;
    }

 	_numEntries = ((uint *)(_file._mmapBase))[0];
	_file._wordOffs = ((uint *)(_file._mmapBase))[1];
	_file._kanaOffs = ((uint *)(_file._mmapBase))[2];
	_file._kanjiOffs = ((uint *)(_file._mmapBase))[3];
	_file._kanaIndexSize = _file._kanjiOffs - _file._kanaOffs;

    if(_entries != 0)
        delete[] _entries;

    _entries = new DictWord[_numEntries];

	uint pos = _file._wordOffs;
	const char *ptr = (const char *)_file._mmapBase;
	ptr += (_file._wordOffs + _numEntries)  * sizeof(uint);

    for(uint i=0; i<_numEntries; i++)
	{
        _entries[i].setPointer((DictWordStorage *)ptr);
		ptr += ((uint *)_file._mmapBase)[pos];
		++pos;
	}

    return true;
}


void Dictionary::J2E(const QString& j, QList<const DictWord> *res)
{
    _cancelled=false;
    bool onlyKana = true;
    bool kanaFirst = false;
    QStringList readings;
    const DictKanji *kanji;
    const DictWord *de;
    int count = 0;
    int step = 0;
	QArray<int> kanjiIndices;

    if(j.isEmpty())
        return;

    QStringList deinfList;
    QCString cstr(j.utf8());
    int cstrlen = cstr.length();

    for(int i=0; i<j.length(); i++)
    {
        if(KanjiDic::isKanji(j[i]))
        {
			kanjiIndices = getKanjiIndices(j[i]);

            onlyKana = false;
            break;
        }
    }

    if(onlyKana && _searchMode == Exact)
    {
		int first;
		int last;
		getKanaRange(j[0], &first, &last);
		if(first == -1)
			return;
        for(int w = first; w <= last; w++)
        {
            de = &_entries[w];
            if(de->reading() == cstr)
            {
                res->append(de);
            }
        }
    }
    else if(onlyKana && _searchMode == Beginning)
    {
		int first;
		int last;
		getKanaRange(j[0], &first, &last);
		if(first == -1)
			return;
        for(int w = first; w <= last; w++)
        {
            de = &_entries[w];
            if(de->reading().left(cstrlen) == cstr)
                res->append(de);
        }
    }
    else if(onlyKana)
    {
        count = 0;

        for(int w = 0; w < entries(); w++)
        {
            if(count++ == 1000)
            {
                emit searchProgress(tr("Searching for %1").arg(j), float(w) / float(entries()));
                qApp->processEvents();
				count=0;
            }


            if(_cancelled)
			{
				_cancelled = false;
                return;
			}
            const DictWord *de = &_entries[w];

            if(_searchMode == Middle)
            {
                if(de->reading().contains(cstr))
                    res->append(de);
            }
            else if(_searchMode == End)
            {
                if(de->reading().right(cstrlen) == cstr)
                    res->append(de);
            }
        }
    }

    else
    {
		count=0;

		for(int w = 0; w < kanjiIndices.count(); w++)
		{
            if(count++ > 100)
            {
                emit searchProgress(tr("Searching for %1").arg(j), float(w) / float(kanjiIndices.count()));
                qApp->processEvents();
				count = 0;
            }

            const DictWord *de = &_entries[kanjiIndices[w]];
            if(_searchMode == Exact)
            {
                if(de->jpn() == cstr)
                    res->append(de);
            }
            if(_searchMode == Beginning)
            {
				if(de->jpn().left(cstrlen) == cstr)
                res->append(de);
            }
            else if(_searchMode == Middle)
            {
                if(de->jpn().contains(cstr))
                res->append(de);
            }
            else if(_searchMode == End)
            {
            	if(de->jpn().right(cstrlen) == cstr)
                res->append(de);
            }

        }
    }

}

void Dictionary::E2J(const QString& e, QList<const DictWord> *res)
{
    int count = 0;

    _cancelled=false;

    if(e.isEmpty())
        return;

    QCString cstr(e.utf8());

    for(int w = 0; w < _numEntries; w++)
    {
        if(count++ == 1000)
        {
         	emit searchProgress(tr("Searching for %1").arg(e), float(w) / float(_numEntries));
            qApp->processEvents();
			count = 0;
        }

        if(_cancelled)
		{
			_cancelled = false;
            return;
		}

        const DictWord *de = &_entries[w];
        if(de->english().contains(cstr,false))
            res->append(de);
    }

    return;
}

void Dictionary::cancelSearch()
{
    _cancelled=true;
}

QArray<int> Dictionary::getKanjiIndices(const QChar& kanji)
{
	QArray<int> ret;

	uint *kanjiIndexBase = (uint *)_file._mmapBase;
	kanjiIndexBase += _file._kanjiOffs / sizeof(uint);
	int kanjiOffs = 0;
	int start = 0;
	int stop;

	while(start == 0)
	{
		if(kanjiIndexBase[kanjiOffs] == kanji.unicode())
		{
			start = kanjiIndexBase[kanjiOffs+1];
			stop = kanjiIndexBase[kanjiOffs+3];
		}
		else
			kanjiOffs += 2;
	}

	ret.resize(stop-start);

	for(int i = start; i < stop; i++)
		ret[i - start] = kanjiIndexBase[i];

	return ret;

}


