/*
    JARDS, Japanese Reference and Drilling System.
    Copyright (C) 2006  Johnny Andersson (e-mail: andersson.johnny@gmail.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <qregexp.h>
#include "item.h"
#include "dictionary.h"
#include "dictmanager.h"
#include "kanjidic.h"

using std::cerr;

Item::Item()
{
}

Box::Box()
{
	_skills = ListDefaults;
	for(int i=0; i < 6; i++)
	{
		_lastAsked[i].setDate(QDate(1970,1,1));
		_rank[i] = 0;
	}
}

Box::~Box()
{
}

int Box::timeLeft(Skills skill) const
{
	if(_rank[skill] == 0)
		return 0;

	int delta = _lastAsked[skill].secsTo(QDateTime::currentDateTime()) / 60;

	return _repeatDelays[_rank[skill] - 1] - delta;

}

QString Box::dateString(Skills skill) const
{
	if(_lastAsked[skill].date().year() == 1970)
		return QObject::tr("N/A");

	return _lastAsked[skill].date().toString();
}

void Box::promote(Skills skill)
{
	_lastAsked[skill] = QDateTime::currentDateTime();
	if(_rank[skill] < 5)
		++_rank[skill];

}

void Box::demote(Skills skill)
{
	_lastAsked[skill] = QDateTime::currentDateTime();
	_rank[skill] = 0;

}



QString Box::saveString() const
{

	QString ret("box=");

	for(int i=0; i < 6; i++)
 		ret += QString("%1%2%3%4%5%6%7,")
			.arg(_lastAsked[i].date().year(),4)
			.arg(_lastAsked[i].date().month(),2)
			.arg(_lastAsked[i].date().day(),2)
			.arg(_lastAsked[i].time().hour(),2)
			.arg(_lastAsked[i].time().minute(),2)
			.arg(_lastAsked[i].time().second(),2)
			.arg(_rank[i],2).replace(QRegExp(" "), "0");

	ret.truncate(ret.length()-1);

	ret += "," + QString().setNum(_skills);

	ret += "\n";

	return ret;

}

void Box::fromString(const QString& str)
{

	QStringList list = QStringList::split(',', str);

	int i=0;

	for(int skill = 0; skill < 6; skill++)
	{
		QString str = list[i];

		_lastAsked[skill].setDate(QDate(str.left(4).toInt(), str.mid(4,2).toInt(), str.mid(6,2).toInt()));
		_lastAsked[skill].setTime(QTime(str.mid(8,2).toInt(), str.mid(10,2).toInt(), str.mid(12,2).toInt()));

		_rank[skill] = str.mid(14, 2).toInt();
		
		i++;
	}

	if(i < list.count())
		_skills = list[i].toInt();	// lists created by jardsmob 1.0 won't have this entry, but we can still be compatible
	else
		_skills = (1 << ListDefaults);

}

int Box::repeatDelay(int box)
{
	if(box < 5)
		return _repeatDelays[box];
	else
		return -1;
}

void Box::setRepeatDelay(int box, int delay)
{
	if(box < 5)
		_repeatDelays[box] = delay;
}

int Kanji::getValue(KanjiValues value) const
{
    if(_dKanji == 0)
        return -1;

    if(value == KanjiFrequency)
        return _dKanji->frequency();
    if(value == KanjiStrokes)
        return _dKanji->strokes();
}

Kanji::Kanji()
{
    _dKanji = 0;
}

Kanji::Kanji(const DictKanji *dKanji)
{
    _dKanji = dKanji;
}

Kanji::~Kanji()
{
}

Item::ItemClasses Kanji::itemClass() const
{
    return KanjiItem;
}

QString Kanji::kanji() const
{
    return QString(_dKanji->character());
}

QString Kanji::kana() const
{
    QString ret;

    ret = QString::fromUtf8(_dKanji->kunReadings()).replace(QRegExp("\\|"),", ");

    if(ret.length() && _dKanji->onReadings().length())
        ret += ", ";

    ret += QString::fromUtf8(_dKanji->onReadings()).replace(QRegExp("\\|"),", ");

    return ret;
}

QString Kanji::english() const
{
    return QString::fromUtf8(_dKanji->meanings()).replace(QRegExp("\\|"),", ");
}

QString Kanji::infoHTML() const
{
    QString str;

    str = QString("<center><largekanji>%1</largekanji></center><br/>").arg(kanji());

    str += "<center><largekana>" + kana() + "</largekana></center><br/>";

    str += "<center><generic>" + english() + "</generic></center><br/>";

    str += "Numerical values:<br/>";

    str += "<ul>";
    str += QString("<li>Freq: %1</li>").arg(getValue(KanjiFrequency));
    str += QString("<li>Strokes: %1</li>").arg(getValue(KanjiStrokes));
    str += "</ul>";

    str += "Radicals:<br/>";

    QString rads = glKanjiDic->getRadicals(_dKanji->character());

    str += "<ul>";

    for(int i=0; i<rads.length(); i++)
    {
        str += "<li><kana>";
        str += rads[i];
        str += "</kana> : " + glKanjiDic->getRadicalMeanings(rads[i]) + QString(", %1 strokes").arg(glKanjiDic->getRadicalStrokes(rads[i]));
        str += "</li>";
    }

    str += "</ul>";


    return str;
}

QString Kanji::listHTML() const
{
    QString ret;

    ret = QString("<largekana><a href=K%1>%2</a></largekana>").arg(kanji()[0]).arg(kanji()[0]) + "<kana> [" + kana() + "] </kana>" + english();

    return ret;
}

QString Kanji::infoString() const
{
    QString ret;

    ret = kanji() + ": " + english();

    return ret;
}


QString Kanji::saveString() const
{
    QString ret;

    ret = "@kanji\n";
    ret += QString("char=%1\n").arg(kanji()[0]);
	ret += _box.saveString();
    return ret;
}

void Kanji::loadStringList(QStringList strings)
{

    for(int i=0; i<strings.count(); i++)
    {
        QString name = QStringList::split('=', strings[i])[0];
        QString value = QStringList::split('=', strings[i])[1];

        if(name == "char")
			_dKanji = glKanjiDic->getDictKanji(value[0]);
		else if(name == "box")
			_box.fromString(value);
	}
}

QStringList Kanji::onReadingsList() const
{
    return _dKanji->onReadingsList();
}

QStringList Kanji::kunReadingsList() const
{
    return _dKanji->kunReadingsList();
}

QStringList Kanji::nameReadingsList() const
{
    return _dKanji->nameReadingsList();
}

QStringList Kanji::meaningsList() const
{
    return _dKanji->meaningsList();
}




Word::Word()
{
}

Word::Word(const DictWord *dword)
{
    _kanji = QString::fromUtf8(dword->jpn());
    _kana = QString::fromUtf8(dword->reading());
    _english = QString::fromUtf8(dword->english()).replace(QRegExp("/"), ", ");
    _markers = dword->markers();
}

Word::~Word()
{
}


Item::ItemClasses Word::itemClass() const
{
    return WordItem;
}

QString Word::kanji() const
{
    return _kanji;
}

QString Word::kana() const
{
    return _kana;
}

QString Word::english() const
{
    return _english;
}

QString Word::infoHTML() const
{
    QString str;

    str += "<center><largekana>" + kanji() + "</largekana>";

    str += "\n<kana> [";
    str += kana();
    str += "]</kana></center>";


    str += "<br/><generic><center>" + english() + "</center></generic><br/><kana>";

    str += markersExplanations();

    str += "</kana><br/>";

    return str;
}

QString Word::listHTML() const
{
    QString str;
    QString src;
    QString tmpstr;

    tmpstr = kanji();

    str += QString("<largekana>");

    for(int i=0; i < tmpstr.length(); i++)
    {
        if(KanjiDic::isKanji(tmpstr[i]))
            str += QString("<a href=K%1>%2</a>").arg(tmpstr[i]).arg(tmpstr[i]);
        else
            str += tmpstr[i];
    }

    str += "<kana><a href=W>";
    str += " [";
    str += kana();
    str += "]";

    str += "</kana></a>";

    str += QString("<generic>");

    str += " (" + markersString() + "): ";

    str += "</generic>";

    tmpstr = english();
    str += tmpstr;


    return str;
}

QString Word::markersString() const
{
    QString ret;
    QStringList list = glDictManager->markers(_markers);

    for(int i=0; i<list.count(); i++)
    {
        ret += list[i] + (i == list.count() -1 ? "" : ", ");
    }

    return ret;
}

QString Word::markersExplanations() const
{
    QString ret;
    QStringList list = glDictManager->markers(_markers);

    ret += "<ul>";

    for(int i=0; i<list.count(); i++)
    {
        ret += "<li>";
        ret += list[i] + ": " + glDictManager->markerDescription(list[i]);
        ret += "</li>";
    }

    ret += "</ul>";

    return ret;
}

QString Word::saveString() const
{
    QString ret;

    ret = "@word\n";

    ret += QString("english=%1\n").arg(_english);
    ret += QString("kanji=%1\n").arg(_kanji);
    ret += QString("kana=%1\n").arg(_kana);
    ret += QString("markers=%1\n").arg(markersNumbers());
	ret += _box.saveString();
    return ret;
}

QString Word::markersNumbers() const
{
    QString str;
    for(int i=0; i<_markers.count(); i++)
    {
        str += QString().setNum(_markers[i]);
        if(i != _markers.count() -1)
            str += ",";
    }

    return str;
}

void Word::markersFromString(const QString& str)
{
    QStringList list = QStringList::split(",",str);
    _markers.resize(list.count());
    for(int i=0; i<list.count(); i++)
        _markers[i] = QString(list[i]).toUShort();
}

void Word::loadStringList(QStringList strings)
{
    for(int i=0; i<strings.count(); i++)
    {
		QStringList list = QStringList::split('=', strings[i]);
        QString name = list[0];
        QString value = list[1];
        if(name == "english")
            _english = value;
        else if(name == "kanji")
            _kanji = value;
        else if(name == "kana")
            _kana = value;
        else if(name == "markers")
            markersFromString(value);
		else if(name == "box")
			_box.fromString(value);
    }

}

QString Word::infoString() const
{
    QString ret;

    ret = kanji() + " [" + kana() + "] " + english();

    return ret;

}


int Box::_repeatDelays[] = {3*24*60,8*60*24,16*60*24,40*60*24,80*60*24};



