/*
    JARDS, Japanese Reference and Drilling System.
    Copyright (C) 2006  Johnny Andersson (e-mail: andersson.johnny@gmail.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef ITEM_H
#define ITEM_H

#include <qstring.h>
#include <qstringlist.h>
#include <qmap.h>
#include <qdatetime.h>
#include "kanjidic.h"

class Box 
{
public:

	enum Skills {KanjiToKana=0, KanjiToMeaning, KanaToKanji, KanaToMeaning, MeaningToKanji, MeaningToKana, ListDefaults};

	Box();
	~Box();

	QDateTime lastAsked(Skills skill) const {return _lastAsked[skill];};
	int timeLeft(Skills skill) const;
	int rank(Skills skill) const {return _rank[skill];};
	void promote(Skills skill);
	void demote(Skills skill);

	QString dateString(Skills skill) const;
	QString saveString() const;
	void fromString(const QString& str);
	int skills() const {return _skills;};
	void setSkills(int skills) {_skills = skills;};

	static int repeatDelay(int box);
	static void setRepeatDelay(int box, int delay);

private:
	int _skills;
	int _rank[6];
	QDateTime _lastAsked[6];
	static int _repeatDelays[5];
	
};

class Item
{
public:

    enum ItemClasses {NoItem, KanjiItem, WordItem};

    Item();

    virtual ~Item() {};

    virtual Item *clone() const {return new Item(*this);};

    virtual ItemClasses itemClass() const {return NoItem;};
    virtual QString kanji() const {return "default kanji";};
    virtual QString kana() const {return "default kana";};
    virtual QString english() const {return "default english";};

    virtual QString infoHTML() const {return "";};
    virtual QString listHTML() const {return "";};
    virtual QString infoString() const {return "";};

    virtual QString saveString() const {return "";};
    virtual void loadStringList(QStringList strings) {};

	void setBox(const Box& box) {_box = box;};

	Box *box() {return &_box;};

protected:
	Box _box;
};


class DictWord;

class Word : public Item
{
public:

    Word();
    Word(const DictWord *dword);
    ~Word();

    Word *clone() const {return new Word(*this);};

    ItemClasses itemClass() const;

    QString kanji() const;
    QString kana() const;
    QString english() const;

    QString infoHTML() const;
    QString listHTML() const;
    QString infoString() const;

    QString saveString() const;
    void loadStringList(QStringList strings);

private:
    void markersFromString(const QString& str);
    QString markersString() const;
    QString markersExplanations() const;
    QString markersNumbers() const;

    QString _kanji;
    QString _kana;
    QString _english;
    QByteArray _markers;

};

class DictKanji;

class Kanji : public Item
{
public:

    enum KanjiValues {KanjiFrequency, KanjiStrokes};

    Kanji();
    Kanji(const DictKanji *kanji);
    ~Kanji();

    Kanji *clone() const {return new Kanji(*this);};

    ItemClasses itemClass() const;

    QString kanji() const;
    QString kana() const;
    QString english() const;

    QString infoHTML() const;
    QString listHTML() const;
    QString infoString() const;

    QStringList onReadingsList() const;
    QStringList kunReadingsList() const;
    QStringList nameReadingsList() const;
    QStringList meaningsList() const;

    QString saveString() const;
    void loadStringList(QStringList strings);

    int getValue(KanjiValues value) const;

private:
    const DictKanji *_dKanji;
};


#endif
