/*
    JARDS, Japanese Reference and Drilling System.
    Copyright (C) 2006  Johnny Andersson (e-mail: andersson.johnny@gmail.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/mman.h>
#include <iostream>
#include <qcstring.h>
#include <qregexp.h>
#include <qtextstream.h>
#include <qeucjpcodec.h>
#include <qstringlist.h>
#include "kanjidic.h"
#include "item.h"

using std::cerr;


DictKanji::DictKanji()
{
}

DictKanji::~DictKanji()
{
}

const CString DictKanji::onReadings() const
{
    return CString((const char *) uint(_ptr) + sizeof(DictKanjiStorage) + _ptr->kunLen, _ptr->onLen);
}

const CString DictKanji::nameReadings() const
{
    return CString((const char *) uint(_ptr) + sizeof(DictKanjiStorage) + _ptr->kunLen + _ptr->onLen, _ptr->nameLen);
}

const CString DictKanji::kunReadings() const
{
    return CString((const char *) uint(_ptr) + sizeof(DictKanjiStorage), _ptr->kunLen);
}

const CString DictKanji::meanings() const
{
    return CString((const char *) uint(_ptr) + sizeof(DictKanjiStorage) + _ptr->kunLen + _ptr->onLen + _ptr->nameLen, _ptr->meaningsLen);
}

const CString DictKanji::radicals() const
{
    return CString((const char *) uint(_ptr) + sizeof(DictKanjiStorage) + _ptr->kunLen + _ptr->onLen + _ptr->nameLen + _ptr->meaningsLen, _ptr->radicalsLen);
}

QStringList DictKanji::onReadingsList() const
{
    return QStringList().split('|', QString::fromUtf8(onReadings()));
}

QStringList DictKanji::kunReadingsList() const
{
    return QStringList().split('|', QString::fromUtf8(kunReadings()));
}

QStringList DictKanji::nameReadingsList() const
{
    return QStringList().split('|', QString::fromUtf8(nameReadings()));
}

QStringList DictKanji::meaningsList() const
{
    return QStringList().split('|', QString::fromUtf8(meanings()));
}


QString KanjiDic::getRadicalMeanings(QChar ch) const
{
    QString *str = _radicals.find(ch.unicode());

    if(str)
    {
        QStringList list;
        list = list.split('|', *str);
        return list[1];
    }
    else
        return "";
}

KanjiDic::KanjiDic()
{
    _entries = 0;
    _numEntries = 0;
    _mmapBase = 0;
    _radicals.setAutoDelete(true);
    _kanjiToRadicals.setAutoDelete(true);
}

KanjiDic::~KanjiDic()
{
    if(_entries != 0)
        delete[] _entries;

    if(_file.isOpen())
    {
        munmap(_mmapBase, _file.size());
        _file.close();
    }

    _radicals.clear();
}

QString KanjiDic::getKanjiStrokes(const Kanji *kanji)
{
    //    return _sd.getStrokes(kanji->strokes(), kanji->jis());
    return "1234";
}

bool KanjiDic::setStrokeDataFile(const QString& fileName)
{
    return _sd.initialize(fileName);
}

bool KanjiDic::setFile(const QString& fileName)
{

    QDataStream s;
    uint temp;
    uint offset;
    uint pos = 0;
    uint mmapSize;

    if(_file.isOpen())
        _file.close();

    if(!_mmapBase == 0)
        munmap(_mmapBase, _file.size());

    _file.setName(fileName);

    if(!_file.open(IO_ReadOnly))
        return false;

    s.setDevice(&_file);

    s.readRawBytes((char *)&_numEntries, sizeof(_numEntries));

    if(_entries != 0)
        delete[] _entries;

    _entries = new DictKanji[_numEntries];

    offset = sizeof(uint) * (_numEntries + 1);    // jump past number_of_entries value + number_of_entries * offset

    _mmapBase = (DictKanjiStorage *) mmap(0, _file.size(), PROT_READ, MAP_SHARED, _file.handle(), 0);

    if(_mmapBase == MAP_FAILED)
    {
        _mmapBase = 0;
        _file.close();
        return false;
    }

    _entries[0].setPointer((DictKanjiStorage *)(uint(_mmapBase) + offset));

    for(uint i=1; i<_numEntries; i++)
    {
        s.readRawBytes((char *)&temp, sizeof(temp));
        pos += temp;
        _entries[i].setPointer((DictKanjiStorage *)(uint(_mmapBase) + offset + pos));
    }

    return true;

}

bool KanjiDic::loadRadicals(const QString& fileName, const QString& codecName)
{

    QString ret;
    QString str;
    QString tmpstr;
    QString meanings;
    QChar rad;
    QTextCodec *codec = 0;
    QFile file;
    QTextStream ts;
    bool ok;
    ushort strokes;
    int p1, p2;

    if(codecName != "utf8")
    {
        codec = QTextCodec::codecForName(codecName);

        if(codec == 0)
            return false;
    }

    file.setName(fileName);

    if(file.open(IO_ReadOnly) == false)
        return false;

    _kanjiToRadicals.clear();

    if(codec)
        ts.setCodec(codec);
    else
        ts.setEncoding(QTextStream::UnicodeUTF8);

    ts.setDevice(&file);

    while(!ts.eof())
    {
        str = ts.readLine();

        if(str[0] == '#')
            continue;

        else if(str[0] == '$')
        {
            rad = str[2];

            tmpstr = str.mid(4, str.find(' ', 4) - 4);
            strokes = (ushort) tmpstr.toInt(&ok,10);

            p1 = str.find("\"");
            p2 = str.find("\"", p1+1);
            meanings = str.mid(p1+1, p2-p1-1).utf8();

            QString *radstr = _strokesToRadicals.find(strokes);
            if(radstr == 0)
                _strokesToRadicals.insert(strokes, new QString(rad));
            else
                radstr->append(rad);
            // lastStrokes=strokes;
            //strokeRadicals = "";

            _radicals.insert(str[2].unicode(), new QString(tmpstr + "|" + meanings));
        }
        else
        {
            for(int i=0; i<str.length(); i++)
            {
                QChar ch = str[i];
                QString *str = _kanjiToRadicals.find(ch.unicode());
                if(str == 0)
                    _kanjiToRadicals.insert(ch.unicode(), new QString(rad));
                else
                    *str += rad;
            }
        }
    }


    return true;
}

QString KanjiDic::getRadicals(QChar kanji)
{
    QString *str = _kanjiToRadicals.find(kanji.unicode());

    if(str)
        return QString(*str);
    else
        return "";
}


QString KanjiDic::getRadicalsByStrokes(int strokes) const
{
    QString *str = _strokesToRadicals.find(strokes);
    if(str)
        return QString(*str);
    else
        return("error");
}

ushort KanjiDic::getRadicalStrokes(QChar ch) const
{
    QString *str = _radicals.find(ch.unicode());
    if(str)
    {
        bool ok;
        QStringList list;
        list = list.split('|', *str);
        return (ushort) list[0].toInt(&ok,10);
    }
    else
        return 0;
}


Kanji KanjiDic::getKanji(QChar ch) const
{
    for(int i=0; i<_numEntries; i++)
    {
        if(_entries[i].unicode() == ch.unicode())
        {
            return Kanji(&_entries[i]);
        }
    }
    return 0;
}

const DictKanji *KanjiDic::getDictKanji(QChar ch) const
{
    for(int i=0; i<_numEntries; i++)
    {
        if(_entries[i].unicode() == ch.unicode())
        {
            return &_entries[i];
        }
    }
    return 0;
}



Kanji KanjiDic::getKanjiByJIS(ushort jis) const
{
    for(int i=0; i<_numEntries; i++)
    {
        if(_entries[i].jis() == jis)
        {
            return Kanji(&_entries[i]);
        }
    }
    return 0;
}

QString KanjiDic::getKanjiByOnReading(const QString& str) const
{
    QString ret;
    for(int i=0; i<_numEntries; i++)
    {
        QStringList list = _entries[i].onReadingsList();
        for(int x=0; x < list.count(); x++)
        {
            if(list[x] == str)
            {
                ret += _entries[i].character();
                break;
            }
        }
    }
    return ret;
}

QString KanjiDic::getKanjiByNameReading(const QString& str) const
{
    QString ret;
    for(int i=0; i<_numEntries; i++)
    {
        QStringList list = _entries[i].nameReadingsList();
        for(int x=0; x <list.count(); x++)
        {
            if(list[x] == str)
            {
                ret += _entries[i].character();
                break;
            }
        }
    }
    return ret;
}

QString KanjiDic::getKanjiByMeaning(const QString& str) const
{
    QString ret;
    for(int i=0; i<_numEntries; i++)
    {
        if((_entries[i].meanings().contains(QRegExp(str))))
            ret += _entries[i].character();
    }
    return ret;
}


QString KanjiDic::getKanjiByStrokes(int strokes, int deviation) const
{
    QString ret;

    for(int i=0; i<_numEntries; i++)
    {
        if((_entries[i].strokes() >= (strokes - deviation)) && (_entries[i].strokes() <= (strokes + deviation)))
        {
            ret += _entries[i].character();
        }
    }
    return ret;
}


QString KanjiDic::getKanjiByRadicals(const QString& rads, int strokes)
{
    if(rads.length() == 0)
        return "";

    QString *str;
    int matchcount = 0;

    QString ret;

    for(int i=0; i < _numEntries; i++)
    {

        if(_entries[i].strokes() != strokes && strokes != 0)
            continue;

        str = _kanjiToRadicals.find(_entries[i].unicode());
        if(str == 0)
            continue;

        for(int c = 0; c < rads.length(); c++)
        {
            if(str->contains(rads[c]))
                matchcount++;
            else
                break;
        }

        if(matchcount == rads.length())
            ret += QChar(_entries[i].unicode());
        matchcount = 0;
    }

    return ret;

}

KanjiDic *glKanjiDic = 0;
