/*
    JARDS, Japanese Reference and Drilling System.
    Copyright (C) 2006  Johnny Andersson (e-mail: andersson.johnny@gmail.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <qpushbutton.h>
#include <qmessagebox.h>
#include <qpopupmenu.h>
#include "fonts.h"
#include "kanjidic.h"
#include "jpnwordentry.h"
#include "jpnlineedit.h"
#include "progressbar.h"

using std::cerr;

JpnWordEntry::JpnWordEntry(QWidget *parent, const char *name, WFlags f)
        : JpnWordEntryWidget(parent, name, f)
{

    connect(__clear, SIGNAL(clicked()), __input, SLOT(clear()));
    connect(__input, SIGNAL(returnPressed()), this, SIGNAL(returnPressed()));
    connect(__input, SIGNAL(keyPressed(int)), this, SLOT(slotKeyPressed(int)));
    connect(__search, SIGNAL(clicked()), this, SIGNAL(returnPressed()));
    connect(__progress, SIGNAL(alertDone()), this, SLOT(slotHideProgressBar()));
    connect(__cancel, SIGNAL(clicked()), this, SLOT(slotCancelClicked()));

	connect(glFonts, SIGNAL(fontsChanged()), this, SLOT(slotFontsChanged()));
	connect(glFonts, SIGNAL(iconsChanged()), this, SLOT(slotIconsChanged()));

    _searchModeMenu = new QPopupMenu(this);
    _exactId = _searchModeMenu->insertItem("Exact");
    _beginningId = _searchModeMenu->insertItem("Beginning");
    _anywhereId = _searchModeMenu->insertItem("Anywhere");
    _endId = _searchModeMenu->insertItem("End");
//    _searchModeMenu->insertItem("Wildcard (*)");
    connect(_searchModeMenu, SIGNAL(activated(int)), this, SLOT(slotSearchModeActivated(int)));
    _searchModeMenu->setItemChecked(_exactId, true);
	__modeMenu->setPopup(_searchModeMenu);

    _inputMenu = new QPopupMenu(this);
    _hiraganaId = _inputMenu->insertItem("Hiragana");
    _katakanaId = _inputMenu->insertItem("Katakana");
    _romajiId = _inputMenu->insertItem("English");
    connect(_inputMenu, SIGNAL(activated(int)), this, SLOT(slotInputModeActivated(int)));
    _inputMenu->setItemChecked(_hiraganaId, true);
    __menu->setPopup(_inputMenu);
	updateMenuCharacter();

	connect(_inputMenu, SIGNAL(activated(int)), this, SLOT(slotInputModeActivated(int)));

    __cancel->hide();
	__progress->hide();
	_currentSearchModeId = _exactId;

	slotSearchModeActivated(_exactId);

}

JpnWordEntry::~JpnWordEntry()
{
}

void JpnWordEntry::slotCancelClicked()
{
	emit cancelClicked();
}

void JpnWordEntry::slotFontsChanged()
{
	__menu->setFont(glFonts->font(Fonts::FONT_Kana));
	__modeMenu->setFont(glFonts->font(Fonts::FONT_Generic));
	__input->setFont(glFonts->font(Fonts::FONT_Kana));
	__progress->setFont(glFonts->font(Fonts::FONT_Kana));
	
}

void JpnWordEntry::slotIconsChanged()
{
	QString str;

	if(glFonts->iconSize() == Fonts::ICONSIZE_Small)	
		str = "small_";
	else
		str = "large_";

	__clear->setPixmap(glFonts->imagePath() + str + "erase.png");	
	__search->setPixmap(glFonts->imagePath() + str + "search.png");	
	__cancel->setPixmap(glFonts->imagePath() + str + "stop.png");	
}

void JpnWordEntry::slotKeyPressed(int key)
{
    if(key == Key_Down)
	{
		if(__input->inputMode() == JpnLineEdit::Hiragana)
			slotInputModeActivated(_katakanaId);
		else if(__input->inputMode() == JpnLineEdit::Katakana)
			slotInputModeActivated(_romajiId);
		else if(__input->inputMode() == JpnLineEdit::Romaji)
			slotInputModeActivated(_hiraganaId);
	}
    else if(key == Key_Up)
	{
		if(_currentSearchModeId == _exactId)
			slotSearchModeActivated(_beginningId);
		else if(_currentSearchModeId == _beginningId)
			slotSearchModeActivated(_anywhereId);
		else if(_currentSearchModeId == _anywhereId)
			slotSearchModeActivated(_endId);
		else if(_currentSearchModeId == _endId)
			slotSearchModeActivated(_exactId);

	}
}

void JpnWordEntry::slotTrimText()
{
    __input->slotTrimText();
}

void JpnWordEntry::slotFocus()
{
    __input->setFocus();
}

void JpnWordEntry::slotFocus(QChar ch)
{
    __input->insert(ch);
    __input->setFocus();
}

void JpnWordEntry::insert(QChar ch)
{
    __input->insert(ch);
}

void JpnWordEntry::slotSetProgress(const QString& message, float progress)
{
	__progress->slotSetText(message, progress);
}

void JpnWordEntry::alert(const QString& message)
{
	__progress->slotSetText(message);
	slotShowProgressBar();
	__cancel->hide();
	__progress->slotAlert();
}

void JpnWordEntry::setBusy(bool busy)
{
    if(busy)
		slotShowProgressBar();
    else
		slotHideProgressBar();
	
}	

void JpnWordEntry::slotHideProgressBar()
{
        __search->show();
		__input->show();
		__menu->show();
		__modeMenu->show();
		__clear->show();
        __cancel->hide();
		__progress->hide();
}

void JpnWordEntry::slotShowProgressBar()
{
        __search->hide();
		__input->hide();
		__modeMenu->hide();
		__menu->hide();
		__clear->hide();
        __cancel->show();
		__progress->show();
}

void JpnWordEntry::slotInputModeActivated(int id)
{
    _inputMenu->setItemChecked(_hiraganaId, false);
    _inputMenu->setItemChecked(_katakanaId, false);
    _inputMenu->setItemChecked(_romajiId, false);
    _inputMenu->setItemChecked(id, true);

    if(id == _hiraganaId)
        __input->setInputMode(JpnLineEdit::Hiragana);
    if(id == _katakanaId)
        __input->setInputMode(JpnLineEdit::Katakana);
    if(id == _romajiId)
        __input->setInputMode(JpnLineEdit::Romaji);

	updateMenuCharacter();
}

void JpnWordEntry::slotSearchModeActivated(int id)
{
    _searchModeMenu->setItemChecked(_beginningId, false);
    _searchModeMenu->setItemChecked(_exactId, false);
    _searchModeMenu->setItemChecked(_endId, false);
    _searchModeMenu->setItemChecked(_anywhereId, false);
    _searchModeMenu->setItemChecked(id, true);

    if(id == _beginningId)
	{
        emit(setSearchMode(Dictionary::Beginning));
		__modeMenu->setText(tr("B"));
	}
    if(id == _exactId)
	{
        emit(setSearchMode(Dictionary::Exact));
		__modeMenu->setText(tr("X"));
	}    
	if(id == _endId)
	{        
		emit(setSearchMode(Dictionary::End));
		__modeMenu->setText(tr("E"));

	}
    if(id == _anywhereId)
	{
        emit(setSearchMode(Dictionary::Middle));
		__modeMenu->setText(tr("A"));
	}
	
	_currentSearchModeId = id;

}

void JpnWordEntry::updateMenuCharacter()
{
	if(__input->inputMode() == JpnLineEdit::Hiragana)
		__menu->setText(QChar(0x3072));
	if(__input->inputMode() == JpnLineEdit::Katakana)
		__menu->setText(QChar(0x30ab));
	if(__input->inputMode() == JpnLineEdit::Romaji)
		__menu->setText("A");

}
